/*
 * atmelio_p.cc
 * 
 * Atmel programmer: Atmel chip parallel method IO routines using latch16bit 
 * for communication. 
 * 
 * Copyright (c) 2003--2005 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "atmelio_p.h"

#include <assert.h>


/* Write optimization. Leave off because switching it on (1) has a bug. 
 * Anyways, the optimization is minimal: Skipping all 0xff when writing 
 * will not result in much advantage... */
#define WRITE_OPT 0

void AtmelIO_Parallel::_ATLoadCommand(uchar cmd)
{
	// Enable command loading: 
	_ATChgControl(
		AT_C_XA1 | AT_C_OE | AT_C_WR,
		AT_C_XA0 | AT_C_BS | AT_C_XTAL );
	_ATDataOutb(cmd);
	_ATPulseXTAL();
}

void AtmelIO_Parallel::_ATSetIdle()
{
	_ATChgControl(AT_C_XA0 | AT_C_XA1, 0);
	_ATChgControl(
		AT_C_XA0 | AT_C_XA1 | AT_C_OE | AT_C_WR,
		AT_C_BS | AT_C_XTAL);
	l16b.WriteSingleByte(DataChannel,0x00);
}

void AtmelIO_Parallel::_ATLoadAddressLow(uchar ladr)
{
	_ATChgControl(0,AT_C_XA0 | AT_C_XA1 | AT_C_BS | AT_C_XTAL);
	_ATDataOutb(ladr);
	_ATPulseXTAL();
}

void AtmelIO_Parallel::_ATLoadAddressHigh(uchar hadr)
{
	_ATChgControl(AT_C_BS,AT_C_XA0 | AT_C_XA1 | AT_C_XTAL);
	_ATDataOutb(hadr);
	_ATPulseXTAL();
}


void AtmelIO_Parallel::_ATLoadDataLow(uchar data)
{
	_ATChgControl(AT_C_XA0,AT_C_BS|AT_C_XA1|AT_C_XTAL);
	_ATDataOutb(data);
	_ATPulseXTAL();
}

void AtmelIO_Parallel::_ATLoadDataHigh(uchar data)
{
	_ATChgControl(AT_C_XA0|AT_C_BS,AT_C_XA1|AT_C_XTAL);
	_ATDataOutb(data);
	_ATPulseXTAL();
}


uchar AtmelIO_Parallel::_ATReadDataOE()
{
	l16b.CtlChannelOutput(DataChannel,/*enable=*/0);
	_ATChgControl(0,AT_C_OE);
	uchar data=l16b.ReadSingleByte(DataChannel,/*disout=*/1);
	_ATChgControl(AT_C_OE,0);
	l16b.CtlChannelOutput(DataChannel,/*enable=*/1);
	return(data);
}


u_int16_t AtmelIO_Parallel::_ATReadWordOE()
{
	// Disable latch16bit data channel output: 
	l16b.CtlChannelOutput(DataChannel,/*enable=*/0);
	
	// Read low byte: 
	_ATChgControl(0,AT_C_OE|AT_C_BS);
	u_int16_t data=(u_int16_t)l16b.ReadSingleByte(DataChannel,/*disout=*/1);
	
	// Read high byte: 
	_ATChgControl(AT_C_BS,0);
	data|=((u_int16_t)l16b.ReadSingleByte(DataChannel,/*disout=*/1))<<8;
	_ATChgControl(AT_C_OE,AT_C_BS);
	
	// Enable latch16bit data channel output again: 
	l16b.CtlChannelOutput(DataChannel,/*enable=*/1);
	
	return(data);
}


// Wait until busy line goes LOW, then HIGH. 
// Has a timeout in case we miss the initial HIGH-LOW transition. 
static size_t busy_misses=0;
void AtmelIO_Parallel::_ATWaitBusy()
{
	unsigned int i=1024;
	// The timeout should be 3us. 
	while((l16b.ReadSingleByte(StatusChannel) & AT_S_RDY) && i)  --i;
	if(!i)
	{  ++busy_misses;  }
	while(!(l16b.ReadSingleByte(StatusChannel) & AT_S_RDY));
}


void AtmelIO_Parallel::_ATWriteDataLow(uchar data)
{
	_ATLoadDataLow(data);
	_ATPulseWR();
	_ATWaitBusy();
}

void AtmelIO_Parallel::_ATWriteDataHigh(uchar data)
{
	_ATLoadDataHigh(data);
	_ATPulseWR();
	_ATWaitBusy();
}


void AtmelIO_Parallel::_ATWriteWord(u_int16_t data)
{
	uchar d=(data & 0xff);
	if(!WRITE_OPT || d!=0xff)  _ATWriteDataLow(d);
	d=(data >> 8);
	if(!WRITE_OPT || d!=0xff)  _ATWriteDataHigh(d);
}

// Like _ATWriteWord() but for chips with PAGEL...
void AtmelIO_Parallel::_ATLatchWord(u_int16_t data)
{
	_ATLoadDataLow(data & 0xff);
	_ATPulsePAGEL();
	_ATLoadDataHigh(data>>8);
	_ATPulsePAGEL();
}


ssize_t AtmelIO_Parallel::ReadFlash(size_t addr,size_t size,u_int16_t *buf)
{
	_ATLoadCommand(AT_CMD_ReadFlash);
	
	size_t chigh=(addr>>8)+1;
	for(size_t a=addr; a<addr+size; a++)
	{
		if((a>>8) != chigh)
		{
			chigh=(a>>8);
			_ATLoadAddressHigh(chigh);
		}
		_ATLoadAddressLow(a & 0xff);
		*(buf++)=_ATReadWordOE();
	}
	
	_ATSetIdle();
	
	return(size);
}


int AtmelIO_Parallel::WriteFlash(size_t addr,size_t size,const u_int16_t *buf)
{
	_ATLoadCommand(AT_CMD_WriteFlash);
	
	busy_misses=0;
	
	if(config.page_size)
	{
		const size_t ps2m1=config.page_size/2-1;
		size_t aend=addr+size;
		for(size_t a=addr; a<aend; a++,buf++)
		{
			_ATLoadAddressLow(a & 0xff);
			_ATLatchWord(*buf);
			if((a&ps2m1)==ps2m1 || a+1==aend)
			{
				// Completed one page. 
				_ATLoadAddressHigh(a>>8);
				// Actually write page: 
				_ATPulseWR();
				_ATWaitBusy();
				// Finish off writing: 
				_ATLoadCommand(AT_CMD_NoOp);
				// Re-load command for writing flash (if needed): 
				if(a+1<aend)
				{  _ATLoadCommand(AT_CMD_WriteFlash);  }
			}
		}
	}
	else
	{
		size_t chigh=(addr>>8)+1;
		size_t aend=addr+size;
		for(size_t a=addr; a<aend; a++,buf++)
		{
			if((a>>8) != chigh)
			{
				chigh=(a>>8);
				_ATLoadAddressHigh(chigh);
			}
			_ATLoadAddressLow(a & 0xff);
			_ATWriteWord(*buf);
		}
	}
	
	_ATSetIdle();
	
	if(busy_misses)
	{  fprintf(stderr,"Note: %u busy misses occured for %u bytes.\n",
		busy_misses,size*2);  }
	
	return(size);
}


ssize_t AtmelIO_Parallel::ReadEEPROM(size_t addr,size_t size,uchar *buf)
{
	_ATLoadCommand(AT_CMD_ReadEEPROM);
	
	size_t chigh=(addr>>8)+1;
	for(size_t a=addr; a<addr+size; a++)
	{
		if((a>>8) != chigh)
		{
			chigh=(a>>8);
			_ATLoadAddressHigh(chigh);
		}
		_ATLoadAddressLow(a & 0xff);
		*(buf++)=_ATReadDataOE();
	}
	
	_ATSetIdle();
	
	return(size);
}


int AtmelIO_Parallel::WriteEEPROM(size_t addr,size_t size,const uchar *buf)
{
	_ATLoadCommand(AT_CMD_WriteEEPROM);
	
	busy_misses=0;
	
	size_t chigh=(addr>>8)+1;
	for(size_t a=addr; a<addr+size; a++,buf++)
	{
		if((a>>8) != chigh)
		{
			chigh=(a>>8);
			_ATLoadAddressHigh(chigh);
		}
		if(!WRITE_OPT || *buf!=0xff)  /* 0xff need not be written. */
		{
			_ATLoadAddressLow(a & 0xff);
			_ATWriteDataLow(*buf);
		}
	}
	
	_ATSetIdle();
	
	if(busy_misses)
	{  fprintf(stderr,"Note: %u busy misses occured for %u bytes.\n",
		busy_misses,size);  }
	
	return(size);
}


int AtmelIO_Parallel::ChipErase()
{
	_ATLoadCommand(AT_CMD_ChipErase);
	
	if(config.has_erase_busy)
	{
		_ATPulseWR();
		_ATWaitBusy();
	}
	else
	{
		// Pulse WR low for >= 15 msec; no activity on busy pin. 
		_ATChgControl(0x00,AT_C_WR);
		usleep(30000);
		_ATChgControl(AT_C_WR,0x00);
	}
	
	_ATSetIdle();
	return(0);
}


int AtmelIO_Parallel::ReadFuseAndLock(uchar *buf)
{
	_ATLoadCommand(AT_CMD_ReadFuseLock);
	int rv=0;
	switch(config.fuse_lock_bytes)
	{
		case 1:
		{
			buf[0]=_ATReadDataOE();
			rv=1;
		}	break;
		case 2:
		{
			u_int16_t tmp=_ATReadWordOE();
			buf[0]=(uchar)(tmp&0xffU);
			buf[1]=(uchar)(tmp>>8);
			rv=2;
		}  break;
		default:  assert(0);  break;
	}
	_ATSetIdle();
	return(rv);
}


AtmelIO_Parallel::ATDevice AtmelIO_Parallel::ReadSignature(Signature *_sig)
{
	Signature *sig=&config.sig;
	
	_ATLoadCommand(AT_CMD_ReadSignature);
	
	_ATLoadAddressLow(0x00);
	sig->_vendor=_ATReadDataOE();
	
	_ATLoadAddressLow(0x01);
	sig->_flash_size=_ATReadDataOE();
	
	_ATLoadAddressLow(0x02);
	sig->_device=_ATReadDataOE();
	
	_ATSetIdle();
	
	// "Parse" the data: 
	sig->ParseSignature(&config);
	
	if(_sig)
	{  *_sig=*sig;  }
	
	return(config.device);
}


void AtmelIO_Parallel::EnterProgMode(volatile int *caught_signal)
{
	fprintf(stderr,"Waiting to enter parallel programming mode... ");
	fflush(stderr);
	
	while(!*caught_signal)
	{
		usleep(1000);
		uchar x=l16b.ReadSingleByte(2);
		fprintf(stderr,"[%02x:%c%c%c]\b\b\b\b\b\b\b\b",x,
			(x & AT_S_PWR) ? 'p' : '-',
			(x & AT_S_PrgMode) ? 'k' : '-',
			(x & AT_S_RDY) ? 'r' : '-');
		if((x & AT_S_PWR) && 
		   (x & AT_S_PrgMode) && 
		   (x & AT_S_RDY) )  break;
	}
	
	if(!*caught_signal) fprintf(stderr,"\a");
	usleep(10000);
	
	fprintf(stderr,*caught_signal ? "[signal]\n" : "[OK\n");
	fflush(stderr);
}


void AtmelIO_Parallel::LeaveProgMode(volatile int *caught_signal)
{
	fprintf(stderr,"Waiting to leave parallel programming mode... ");
	fflush(stderr);
	
	while(!*caught_signal)
	{
		usleep(1000);
		uchar x=l16b.ReadSingleByte(2);
		if(!(x & AT_S_PrgMode))  break;
	}
	
	if(!*caught_signal) fprintf(stderr,"\a");
	
	fprintf(stderr,*caught_signal ? "[signal]\n" : "OK\n");
	fflush(stderr);
}


int AtmelIO_Parallel::Connect()
{
	// Disable channel A,B input: 
	l16b.SelectInputChannel(/*channel=*/0);
	// Set all input channels transparent: 
	l16b.CtlInChannelTransparent(/*channel=*/3,/*transparent=*/1);
	
	// Set all output channels non-transparent: 
	l16b.CtlOutChannelTransparent(/*channel=*/3,/*transparent=*/0);
	
	// Put some (sane) bits into the latches: 
	l16b.WriteSingleByte(/*channel=*/DataChannel,0x00);
	control_bits=(AT_C_OE|AT_C_WR|AT_C_XA0|AT_C_XA1);
	l16b.WriteSingleByte(/*channel=*/ControlChannel,control_bits);
	
	// Enable channel A,B output (after 
	// the WriteSingleByte() calls above): 
	l16b.CtlChannelOutput(/*channel=*/3,/*enable=*/1);
	
	return(0);
}

int AtmelIO_Parallel::Disconnect()
{
	// Disable channel A,B input: 
	l16b.SelectInputChannel(/*channel=*/0);
	// Set all input channels non-transparent: 
	l16b.CtlInChannelTransparent(/*channel=*/3,/*transparent=*/0);
	
	// Set all output channels non-transparent: 
	l16b.CtlOutChannelTransparent(/*channel=*/3,/*transparent=*/0);
	
	// Finally, put something sane on the output: 
	control_bits=(AT_C_OE|AT_C_WR|AT_C_XA0|AT_C_XA1);
	l16b.WriteSingleByte(/*channel=*/ControlChannel,control_bits);
	l16b.WriteSingleByte(/*channel=*/DataChannel,0x00);
	
	// Disable channel A,B output: 
	l16b.CtlChannelOutput(/*channel=*/3,/*enable=*/0);
	
	return(0);
}


AtmelIO_Parallel::AtmelIO_Parallel(LPPortIO *pio,uchar strobe_line,uchar ctlsel_line) : 
	AtmelIO(pio,strobe_line,ctlsel_line),
	DataChannel(1),ControlChannel(2),StatusChannel(2)
{
	control_bits=0x00;
	Disconnect();
}

AtmelIO_Parallel::~AtmelIO_Parallel()
{
	Disconnect();
}
