/*
 * atmelio_s.cc
 * 
 * Atmel programmer: Atmel chip serial method IO routines using latch16bit 
 * for communication. 
 * 
 * Copyright (c) 2004--2005 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "atmelio_s.h"

#include <assert.h>

// Only use delay timing any no data polling?
#define FORCE_DELAY_TIMING 0


extern volatile int caught_signal;


uchar AtmelIO_Serial::_ATSerialIO(u_int32_t cmd)
{
	register u_int32_t mask=0x80000000U;
	for(; mask!=0x00000080U; mask>>=1)
	{
		register uchar bitval=(cmd & mask) ? AT_C_MOSI : 0x00;
		DataOutb(bitval);
		DataOutb(bitval | AT_C_SCLK);
	}
	uchar response=0U;
	for(; mask; mask>>=1)
	{
		register uchar bitval=(cmd & mask) ? AT_C_MOSI : 0x00;
		DataOutb(bitval);
		DataOutb(bitval | AT_C_SCLK);
		response<<=1;
		uchar resp=l16b.ReadSingleByte(InChannel);
		if(resp & AT_S_MISO)  response|=1U;
	}
	DataOutb(0x00);
	return(response);
}


u_int32_t AtmelIO_Serial::_ATSerialIO32(u_int32_t cmd)
{
	u_int32_t response=0U;
#if 1  /* Normal, faster routine. */
	for(int bit=31; bit>=0; bit--)
	{
		uchar bitval=(cmd & (1U<<bit)) ? AT_C_MOSI : 0x00;
		DataOutb(bitval);
		DataOutb(bitval | AT_C_SCLK);
		response<<=1;
		uchar resp=l16b.ReadSingleByte(InChannel);
		if(resp & AT_S_MISO)  response|=1U;
	}
#else  /* No advantage here: */
	for(int bit=31; bit>=0; bit--)
	{
		uchar bitval=(cmd & (1U<<bit)) ? AT_C_MOSI : 0x00;
		DataOutb(bitval);
		response<<=1;
		uchar resp=l16b.ReadSingleByte(InChannel);
		if(resp & AT_S_MISO)  response|=1U;
		DataOutb(bitval | AT_C_SCLK);
		DataOutb(bitval);
	}
#endif
	DataOutb(0x00);
	return(response);
}


ssize_t AtmelIO_Serial::ReadFlash(size_t addr,size_t size,u_int16_t *buf)
{
	for(size_t a=addr; a<addr+size; a++)
	{
		uchar low =  _ATSerialIO( AT_CMD_ReadFlashL | (u_int32_t(a)<<8) );
		uchar high = _ATSerialIO( AT_CMD_ReadFlashH | (u_int32_t(a)<<8) );
		*(buf++) = (u_int16_t(high)<<8) | u_int16_t(low);
		if(caught_signal)  return(a-addr);
	}
	return(size);
}


int AtmelIO_Serial::_PollFlashOrWait_NoPage(u_int32_t cmd, uchar xbyte)
{
#if !FORCE_DELAY_TIMING
	// At least 0x7f is special and cannot be polled. 
	if(xbyte==0x7f)
#endif
	{
		// AT90S2313 wants 6ms at 4V and 4ms at 5V. 
		usleep(6000);
		return(0);
	}
	
#if !FORCE_DELAY_TIMING
	// Data polling. Simply read the location until the returned 
	// value is what we expect. 
	uchar curr;
	for(int ntry=0; ntry<2000; ntry++)
	{
		curr = _ATSerialIO( cmd );
		if(curr==xbyte || caught_signal)  return(0);
	}
	
	fprintf(stderr,"SPI: Flash addr=0x%x: write timeout or "
		"failure (0x%02x/0x%02x)\n",(cmd & 0x00ffff00U)>>8,curr,xbyte);
	return(1);
#endif
}


int AtmelIO_Serial::WriteFlash(size_t addr,size_t size,const u_int16_t *buf)
{
	if(config.page_size)
	{
		const size_t ps2m1=config.page_size/2-1;
		size_t aend=addr+size;
#if !FORCE_DELAY_TIMING
		u_int32_t poll_cmd=0U;
		uchar poll_byte=0;
#endif
		for(size_t a=addr; a<aend; a++,buf++)
		{
			u_int32_t addr_cmd = (u_int32_t(a)<<8);
			
			uchar xbyte = uchar(*buf & 0xffU);
#if !FORCE_DELAY_TIMING
			if(!poll_cmd && xbyte!=0xff && xbyte!=0x7f && xbyte!=0x80)  // ..be sure
			{
				poll_cmd = AT_CMD_ReadFlashL | addr_cmd;
				poll_byte = xbyte;
			}
#endif
			_ATSerialIO( AT_CMD_WriteFlashL | addr_cmd | u_int32_t(xbyte) );
			
			xbyte = uchar(*buf >> 8);
#if !FORCE_DELAY_TIMING
			if(!poll_cmd && xbyte!=0xff && xbyte!=0x7f && xbyte!=0x80)  // ..be sure
			{
				poll_cmd = AT_CMD_ReadFlashH | addr_cmd;
				poll_byte = xbyte;
			}
#endif
			_ATSerialIO( AT_CMD_WriteFlashH | addr_cmd | u_int32_t(xbyte) );
			
			if((a & ps2m1)==ps2m1 || a+1==aend)
			{
				// Completed one page. 
				_ATSerialIO( AT_CMD_WritePage | (u_int32_t(a)<<8) );
				
#if !FORCE_DELAY_TIMING
				if(poll_cmd)
				{
					for(int ntry=0; ntry<8000; ntry++)
					{
						xbyte = _ATSerialIO( poll_cmd );
						if(xbyte==poll_byte)
						{  goto okay;  }
					}
					fprintf(stderr,"SPI: Flash page=0x%x: write timeout or "
						"failure (0x%02x/0x%02x)\n",
						a & 0xffffffc0,xbyte,poll_byte);
					return(a-addr);
					okay:;
				}
				else
#endif
				{
					// ATmega161 wants 14ms for a flash page. 
					usleep(15000);
				}
				
				if(caught_signal)  return(a-addr);
				
#if !FORCE_DELAY_TIMING
				poll_cmd=0U;
				poll_byte=0;
#endif
				
				//fprintf(stderr,"Programmed page 0x%x. Press enter...",a>>6);
				//getchar();
			}
		}
	}
	else
	{
		size_t aend=addr+size;
		for(size_t a=addr; a<aend; a++,buf++)
		{
			u_int32_t addr_cmd = (u_int32_t(a)<<8);
			
			uchar xbyte = uchar(*buf & 0xffU);
			_ATSerialIO( AT_CMD_WriteFlashL | addr_cmd | u_int32_t(xbyte) );
			
			if(_PollFlashOrWait_NoPage( AT_CMD_ReadFlashL | addr_cmd, xbyte ))
			{  return(a-addr);  }
			
			if(caught_signal)  return(a-addr);
			
			xbyte = uchar(*buf >> 8);
			_ATSerialIO( AT_CMD_WriteFlashH | addr_cmd | u_int32_t(xbyte) );
			
			if(_PollFlashOrWait_NoPage( AT_CMD_ReadFlashH | addr_cmd, xbyte ))
			{  return(a-addr);  }
			
			if(caught_signal)  return(a-addr);
		}
	}
	
	return(size);
}


ssize_t AtmelIO_Serial::ReadEEPROM(size_t addr,size_t size,uchar *buf)
{
	for(size_t a=addr; a<addr+size; a++)
	{
		*(buf++) = _ATSerialIO( AT_CMD_ReadEEPROM | (u_int32_t(a)<<8) );
		if(caught_signal)  return(a-addr);
	}
	return(size);
}


int AtmelIO_Serial::WriteEEPROM(size_t addr,size_t size,const uchar *buf)
{
	for(size_t a=addr; a<addr+size; a++,buf++)
	{
		u_int32_t addr_cmd = (u_int32_t(a)<<8);
		// First, read the current content and see: 
		uchar curr = _ATSerialIO( AT_CMD_ReadEEPROM | addr_cmd );
		// Need not write if the content will not change: 
		if(curr==*buf)  continue;
		// Write the content: 
		_ATSerialIO( AT_CMD_WriteEEPROM | addr_cmd | u_int32_t(*buf) );
		// Okay, data polling is nice, of course, but different devices 
		// require different data polling algorithms: 
		// ATmega161 returns 0xff until done. 
		// AT90S2313 returns 0x80 during erase and 0x7f during write. 
		// So, for now, I use delay timing for these 3 values. 
#if !FORCE_DELAY_TIMING
		switch(*buf)
		{
			case 0x7f:
			case 0x80:
			case 0xff:
#endif
				// ATmega161 requires 3.4ms. 
				// AT90S2313 needs 6ms at 4V and 4ms at 5V
				// ...so we should be on the safe side with 6ms. 
				// (The actual delay depends on the time resolution and 
				// will normally be 10ms on my P-100.)
				usleep(6000);
#if !FORCE_DELAY_TIMING
				break;
			default:
				// Data polling. Simply read the location until the returned 
				// value is what we expect. 
				for(int ntry=0; ntry<2000; ntry++)
				{
					curr = _ATSerialIO( AT_CMD_ReadEEPROM | addr_cmd );
					if(curr==*buf || caught_signal)
					{  goto okay;  }
				}
				fprintf(stderr,"SPI: EEPROM addr=0x%x: write timeout or "
					"failure (0x%02x/0x%02x)\n",a,curr,*buf);
				return(a-addr);
				okay:;
				break;
		}
#endif
		
		if(caught_signal)  return(a-addr);
	}
	
	return(size);
}


int AtmelIO_Serial::ChipErase()
{
	_ATSerialIO( AT_CMD_ChipErase );
	
	// Using sleep timing here in any case. 
	// ATMega161 wants 28ms here, AT90S2313 only <=12ms at >=4V. 
	usleep(30000);
	
	fprintf(stderr,"SPI: Chip erased. Give RESET a positive pulse.\n");
	return(1);
}


int AtmelIO_Serial::ReadFuseAndLock(uchar *buf)
{
	switch(config.fuse_lock_bytes)
	{
		case 1:
			// The only supported device here is the AT90S2313 which 
			// does not allow to read fuse and lock bits via SPI. 
			fprintf(stderr,"SPI: Fuse and lock bits not readable via SPI.\n");
			return(0);
		case 2:
			// ATMega161 has no trouble. 
			buf[0] = _ATSerialIO(AT_CMD_ReadFuse);
			buf[1] = _ATSerialIO(AT_CMD_ReadLock);
			return(2);
		case 3:
			// ATMega8515, e.g. 
			buf[0] = _ATSerialIO(AT_CMD_ReadFuse);
			buf[1] = _ATSerialIO(AT_CMD_ReadLock);
			buf[2] = _ATSerialIO(AT_CMD_ReadFuseH);
			return(3);
		default:  assert(0);  break;
	}
}


void AtmelIO_Serial::WriteFuse(uchar val,char is_high)
{
	u_int32_t cmd = is_high ? AT_CMD_WriteFuseH : AT_CMD_WriteFuse;
	
	cmd|=(u_int32_t)val;
	
	_ATSerialIO(cmd);
}


AtmelIO_Serial::ATDevice AtmelIO_Serial::ReadSignature(Signature *_sig)
{
	Signature *sig=&config.sig;
	
	//fprintf(stderr,"RESP= 0x%x 0x%x 0x%x\n",
	//	_ATSerialIO32( AT_CMD_ReadSig | (0x0U << 8) ),
	//	_ATSerialIO32( AT_CMD_ReadSig | (0x1U << 8) ),
	//	_ATSerialIO32( AT_CMD_ReadSig | (0x2U << 8) ) );
	
	sig->_vendor=    _ATSerialIO( AT_CMD_ReadSig | (0x0U << 8) );
	sig->_flash_size=_ATSerialIO( AT_CMD_ReadSig | (0x1U << 8) );
	sig->_device=    _ATSerialIO( AT_CMD_ReadSig | (0x2U << 8) );
	
	// "Parse" the data: 
	sig->ParseSignature(&config);
	
	if(_sig)
	{  *_sig=*sig;  }
	
	return(config.device);
}


void AtmelIO_Serial::EnterProgMode(volatile int *_caught_signal)
{
	fprintf(stderr,"Waiting to enter serial programming mode (press enter)... ");
	fflush(stderr);
	
	// Write 0 and enable output: 
	l16b.WriteSingleByte(OutChannel,0x00);
	l16b.CtlChannelOutput(OutChannel,/*enable=*/1);
	l16b.CtlOutChannelTransparent(OutChannel,/*transparent=*/1);
	
	// Okay, now we need to send the programming mode enable command. 
	// And we need to get in sync with the device...
	for(;;)
	{
		getchar();
		if(*_caught_signal)
		{  fprintf(stderr,"[signal]\n");  return;  }
		
		usleep(30000);
		
		fprintf(stderr,"Enabling SPI programming: ");
		fflush(stderr);
		
		// Send the programming enable instruction over the link 
		// and read the echoed data. 
		u_int32_t response=_ATSerialIO32(AT_CMD_ProgEn);
		
		if(*_caught_signal)
		{  fprintf(stderr,"[signal]\n");  break;  }
		
		if((response&0xffffU)==((AT_CMD_ProgEn>>8)&0xffffU))
		{
			fprintf(stderr,"OK [0x%08x]\n",response);
			fflush(stderr);
			break;
		}
		
		// Make a single positive SCLK pulse: 
		// AT90S2313 data sheet states that we can also make a single 
		// clock pulse and try again to see if we're then in sync. 
		// However, ATMega161's data sheet says we should take reset righ 
		// shortly and try again. 
		// Hence, I make one pulse here. If the user does the reset pulse 
		// we're back anyways. If (s)he just hits enter without applying 
		// the reset pulse, we're trying the AT90S2313-method... (which is 
		// risky in case the cycled/shifted command collides with some 
		// other command and we already enabled programming). 
		DataOutb(AT_C_SCLK);
		DataOutb(0x00);
		
		fprintf(stderr,"failed [0x%08x]; do H reset pulse & press enter ",
			response);
		fflush(stderr);
	}
	
	// We wait a little here to be sure. Should be pretty useless, though. 
	usleep(10000);
}


void AtmelIO_Serial::LeaveProgMode(volatile int * /*_caught_signal*/)
{
	// Write 0 and disable output: 
	l16b.CtlChannelOutput(OutChannel,/*enable=*/0);
	l16b.WriteSingleByte(OutChannel,0x00);
	
	fprintf(stderr,"Serial programming done (left programming mode).\n");
	fflush(stderr);
}


int AtmelIO_Serial::Connect()
{
	// Disable channel A,B input: 
	l16b.SelectInputChannel(/*channel=*/0);
	// Set input channel non-transparent: 
	l16b.CtlInChannelTransparent(InChannel,/*transparent=*/0);
	
	// Set all output channel non-transparent: 
	l16b.CtlOutChannelTransparent(OutChannel,/*transparent=*/0);
	
	// Put some (sane) bits into the latche: 
	l16b.WriteSingleByte(OutChannel,0x00);
	
	// Disable output (after the WriteSingleByte() calls above): 
	l16b.CtlChannelOutput(OutChannel,/*enable=*/0);
	
	return(0);
}

int AtmelIO_Serial::Disconnect()
{
	// Disable channel A,B input: 
	l16b.SelectInputChannel(/*channel=*/0);
	// Set input channel non-transparent: 
	l16b.CtlInChannelTransparent(InChannel,/*transparent=*/0);
	
	// Set all output channel non-transparent: 
	l16b.CtlOutChannelTransparent(OutChannel,/*transparent=*/0);
	
	// Put some (sane) bits into the latche: 
	l16b.WriteSingleByte(OutChannel,0x00);
	
	// Disable output (after the WriteSingleByte() calls above): 
	l16b.CtlChannelOutput(OutChannel,/*enable=*/0);
	
	return(0);
}


AtmelIO_Serial::AtmelIO_Serial(LPPortIO *pio,uchar strobe_line,uchar ctlsel_line) : 
	AtmelIO(pio,strobe_line,ctlsel_line),
	InChannel(1),OutChannel(1)
{
	Disconnect();
}

AtmelIO_Serial::~AtmelIO_Serial()
{
	Disconnect();
}
