/*
 * atmelio_s.h
 * 
 * Atmel programmer: Atmel chip IO routines using latch16bit for 
 * communication: Serial transfer method. 
 * 
 * Copyright (c) 2004--2005 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _ATMELPRG_ATMELIO_SERIAL_H_
#define _ATMELPRG_ATMELIO_SERIAL_H_

#include "atmelio.h"


// This extra delay is needed on my Pentium 100 when using a system clock 
// of 1MHz: 
// The SCLK pulses will be nearly 2.5us which is above the required 2us. 
#define SERIAL_DELAY  for(unsigned short int j=0; j<24; j++);


class AtmelIO_Serial : public AtmelIO
{
	public:
		// ATMEL commands: 
		// 0 = LOW, 1 = HIGH, x = don't care, a,b = address high,low, 
		// i,o = data input,output
		// Note: For devices with PEGEL (page load), the WriteFlashL,H 
		//       is actually the load command and the real _write_ command 
		//       is WritePage. 
		static const u_int32_t 
		AT_CMD_ProgEn      = 0xAC530000U, // 10101100 01010011 xxxxxxxx xxxxxxxx
		AT_CMD_ChipErase   = 0xAC800000U, // 10101100 100xxxxx xxxxxxxx xxxxxxxx
		AT_CMD_ReadFlashL  = 0x20000000U, // 00100000 aaaaaaaa bbbbbbbb oooooooo
		AT_CMD_ReadFlashH  = 0x28000000U, // 00101000 aaaaaaaa bbbbbbbb oooooooo
		AT_CMD_ReadEEPROM  = 0xA0000000U, // 10100000 aaaaaaaa bbbbbbbb oooooooo
		AT_CMD_WriteFlashL = 0x40000000U, // 01000000 aaaaaaaa bbbbbbbb iiiiiiii
		AT_CMD_WriteFlashH = 0x48000000U, // 01001000 aaaaaaaa bbbbbbbb iiiiiiii
		AT_CMD_WritePage   = 0x4C000000U, // 01001100 aaaaaaaa bbxxxxxx iiiiiiii
		AT_CMD_WriteEEPROM = 0xC0000000U, // 11000000 aaaaaaaa bbbbbbbb iiiiiiii
		AT_CMD_ReadLock    = 0x58000000U, // 01011000 xxxxxxxx xxxxxxxx oooooooo
		AT_CMD_WriteLock   = 0xACE0FF00U, // 10101100 111xxiix xxxxxxxx iiiiiiii
		AT_CMD_ReadFuse    = 0x50000000U, // 01010000 00000000 xxxxxxxx oooooooo
		AT_CMD_ReadFuseH   = 0x58080000U, // 01011000 00001000 xxxxxxxx oooooooo
		AT_CMD_WriteFuse   = 0xACA0FF00U, // 10101100 10100000 xxxxxxxx iiiiiiii
		AT_CMD_WriteFuseH  = 0xACA8FF00U, // 10101100 10101000 xxxxxxxx iiiiiiii
		AT_CMD_ReadSig     = 0x30000000U, // 00110000 xxxxxxxx xxxxxxbb oooooooo
		AT_CMD_ReadCalib   = 0x38000000U; // 00111000 00xxxxxx 00000000 oooooooo
		
		// NOTE: When WRITING serial data, data is clocked on the RISING 
		//       edge of the clock; when READING, it is clocked on the 
		//       FALLING edge of the clock. 
		// Hence: - Put data on MOSI
		//        - Take SCLK high
		//        - Read data from MISO
		//        - Take SCLK low
		// And for the connection, send MSB first, LSB last. 
		
		// ATMEL pins (output; "control")
		enum
		{
			AT_C_MOSI= 0x01,
			AT_C_SCLK= 0x02,
		};
		// ATMEL pins (input; "status")
		enum
		{
			AT_S_MISO=   0x01,
		};
		
	private:
		// Latch16Bit board is used for communication: 
		//Latch16BitIO l16b;  <-- See base class. 
		
		// Some config: 
		const int InChannel;
		const int OutChannel;
		
		// Some info about the chip we're connected to: 
		//ATConfig config;  <-- See base class. 
		
		// A wrapper for DataOutb() but with delay. 
		inline void DataOutb(register uchar x)
		{
			l16b.DataOutb(x);
			SERIAL_DELAY
		}
		
		// Master serial IO function: Sends 4-byte command passed in cmd 
		// and returns last echoed byte (often some response). 
		uchar _ATSerialIO(u_int32_t cmd);
		// This function returns the complete 32bit response (slower): 
		u_int32_t _ATSerialIO32(u_int32_t cmd);
		
		// For non-paged devices: do flash polling. 
		int _PollFlashOrWait_NoPage(u_int32_t cmd, uchar xbyte);
		
	public:
		AtmelIO_Serial(LPPortIO *pio,uchar strobe_line,uchar crlsel_line);
		~AtmelIO_Serial();
		
		// [overriding virtuals]
		int Connect();
		int Disconnect();
		void EnterProgMode(volatile int *caught_signal);
		void LeaveProgMode(volatile int *caught_signal);
		ATDevice ReadSignature(Signature *sig=NULL);
		int ReadFuseAndLock(uchar *fuse_and_lock);
		int ChipErase();
		ssize_t ReadFlash(size_t addr,size_t size,u_int16_t *buf);
		int WriteFlash(size_t addr,size_t size,const u_int16_t *buf);
		ssize_t ReadEEPROM(size_t addr,size_t size,uchar *buf);
		int WriteEEPROM(size_t addr,size_t size,const uchar *buf);
		void WriteFuse(uchar val,char is_high);
};

#endif  /* _ATMELPRG_ATMELIO_SERIAL_H_ */
