/*
 * buffer.h
 * 
 * Header for the class DataBuffer, representing the buffers used in 
 * the Atmel programmer. 
 * 
 * Copyright (c) 2003--2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _BUFFER_H_
#define _BUFFER_H_

#include <sys/types.h>
#include "atmelio.h"


struct DataBuffer
{
	// There is a list of data buffers. 
	DataBuffer *next;
	
	// Buffer content: 
	char *buf;
	size_t size;
	size_t allocsize;
	
	// Name of the buffer: 
	char *name;
	
	DataBuffer(const char *name);  // name is copied
	~DataBuffer();
	
	// Clear buffer content: 
	void Clear();
	
	// Call Clear() and then allocate specified amount of memory. 
	void Alloc(size_t size);
	
	// Read in binary file: 
	// Return value: 
	//  0 -> OK
	// -2 -> open failed (error written)
	// -3 -> read failed (error written)
	int ReadBinaryFile(const char *file);
	
	// Read/write flash from/to Atmel device: 
	// adr: in 2byte words; len in BYTES!
	// Return value: 0
	int ReadFlash(AtmelIO *atio,size_t adr,size_t len);
	int WriteFlash(AtmelIO *atio,size_t adr);
	
	// Read/write EEPROM from/to Atmel device: 
	// adr, len in BYTES. 
	// Return value: 0
	int ReadEEPROM(AtmelIO *atio,size_t adr,size_t len);
	int WriteEEPROM(AtmelIO *atio,size_t adr);
	
	// Write content to specified binary file. 
	// Rerturn value: 
	//  0 -> OK
	// -2 -> open failed (error written)
	// -3 -> write failed (error written)
	int WriteBinaryFile(const char *file);
	
	// Write hex dump of buffer to passed stream (stdout normally). 
	void HexDumpBuffer(FILE *out,int with_ascii=0);
	
	void *operator new(size_t size);
	void operator delete(void *ptr);
};


class DataBufferList
{
	private:
		DataBuffer *head;
	public:
		DataBufferList();
		~DataBufferList();
		
		// Get a buffer with the specified name. 
		// If do_create is set, create a new one if the buffer 
		// is not found, else return NULL. 
		DataBuffer *GetDataBuffer(const char *name,int do_create);
};

#endif  /* _BUFFER_H_ */
