/*
 * main.cc
 * 
 * Atmel programmer using latch16bit -- main file. 
 * 
 * Copyright (c) 2003--2005 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>

#include "buffer.h"
#include "atmelio_p.h"
#include "atmelio_s.h"

#include <assert.h>

#define VERSION "2.1"

// Parallel port base address: 
const size_t parport_baseaddr=0x378;


static void PrintHelp()
{
	printf(
		"USAGE: atmelprg [--help|--version] [-command...]\n"
		"       atmelspi [--help|--version] [-command...]\n"
		"The program simply executes the commands you specify in the\n"
		"order as they appear on the command line. Each command begins\n"
		"with a dash and has a different number of (following) arguments.\n"
		"Most commands operate on \"buffers\". Buffers are identified by\n"
		"their name which is given to them during creation.\n"
		"Commands:\n"
		"  -readsig\n"
		"      read signature of Atmel device (automatically done\n"
		"      before first device access)\n"
		"  -readfuse\n"
		"      read fuse and lock bits\n"
		"  -load file [buffer]\n"
		"      load specified binary file into buffer with specfied name\n"
		"      (defaults to file name w/o path)\n"
		"  -save buffer file\n"
		"      save specified buffer to binary file\n"
		"  -dump buffer\n"
		"  -dumpasc buffer\n"
		"      hex dump buffer to stdout, -dumpasc includes ASCII line\n"
		"      but requires more than 80 comumns\n"
		"  -erase\n"
		"      perform chip erase which is required before writing flash\n"
		"      (SPI allows writing EEPROM without erase)\n"
		"      [Some devices can be programmed without erase (ATMega161) resulting\n"
		"       in new=~(~old | ~data) [I believe...].]\n"
		"  -readflash [buffer]\n"
		"      read flash content and save in buffer (default name: \"flash\")\n"
		"  -writeflash [buffer]\n"
		"      write content from buffer to device's flash memory\n"
		"      (default buffer name: \"flash\")\n"
		"  -readeeprom [buffer]\n"
		"      read EEPROM content and save in buffer (default name: \"eeprom\")\n"
		"  -writeeeprom [buffer]\n"
		"      write content from buffer to device's EEPROM memory\n"
		"      (default buffer name: \"eeprom\")\n"
		"  -compare bufA bufB\n"
		"      compare two buffers and count differences (verify data)\n"
		"  -special ...\n"
		"      execute special command compiled into this program; see source code\n"
		"Supported devices: AT90S2313, ATMega161; spi: ATMega8515; untested: AT90S4433\n"
		"                   (more can be added easily)\n"
		"NOTE: As of version 2.0, atmelprg supports serial (SPI) and parallel\n"
		"      downloading. For serial downloading, you must call the program\n"
		"      atmelspi, for parallel downloading use atmelprg. (Use a symlink.)\n"
		"atmelprg version %s (c) 2003--2005 by Wolfgang Wieser\n"
		"         This program comes with absolutely no warranty and may be\n"
		"         copied under the terms of the GNU General Public License\n",VERSION);
}


// Default C++ (de)allocation: 
void *operator new(size_t size)
{
	void *ptr=malloc(size);
	assert(ptr);
	return(ptr);
}
void operator delete(void *ptr)
{
	if(ptr)  free(ptr);
}


// Download mode: 
enum DownloadMode
{
	DL_Unknown=0,
	DL_Parallel=1,
	DL_Serial=2,
} downloadmode=DL_Unknown;

// Initialized early in main(): 
DataBufferList *dblist;

// This is initialized when we first need it. 
LPPortIO *ppio=NULL;
AtmelIO *atio=NULL;  // AtmelIO_Parallel or AtmelIO_Serial

// Must be valid when in programming mode: 
AtmelIO::Signature dev_sig;

// Modes: 0 -> idle mode; 1 -> programming
int current_mode=0;

volatile int caught_signal=0;

void catch_sig(int /*sig*/)
{
	++caught_signal;
}


void InstallSignalHandler()
{
	struct sigaction act;
	sigemptyset(&act.sa_mask);
	act.sa_handler=&catch_sig;
	act.sa_flags=0;
	act.sa_restorer=NULL;
	
	sigaction(SIGTERM,&act,NULL);
	sigaction(SIGINT,&act,NULL);
}


// Always returns 0 and 1 on failure. 
static int EnsureProgramMode(int reread_sig=0)
{
	if(current_mode==1 && !reread_sig)
	{
		// Seems we're already in programming mode. 
		return(0);
	}
	
	if(!ppio)
	{  ppio=new LPPortIO(parport_baseaddr,NULL);  }
	
	if(!atio)
	{
		switch(downloadmode)
		{
			case DL_Parallel:
				atio=new AtmelIO_Parallel(ppio,
					LPPortIO::C_STROBE,LPPortIO::C_AUTOLF);
				break;
			case DL_Serial:
				atio=new AtmelIO_Serial(ppio,
					LPPortIO::C_STROBE,LPPortIO::C_AUTOLF);
				break;
			default:  assert(0);
		}
	}
	
	if(!current_mode)
	{
		atio->Connect();
		atio->EnterProgMode(&caught_signal);
		if(caught_signal)  return(1);
	}
	
	AtmelIO::ATDevice dev=atio->ReadSignature(&dev_sig);
	if(dev==AtmelIO::AD_Unknown)
	{  fprintf(stderr,"Failed to detect device (0x%02x%02x%02x).\n",
		dev_sig._vendor,dev_sig._flash_size,dev_sig._device);  return(1);  }
	else
	{
		char psize[24];
		if(atio->GetConfig()->page_size)
		{  snprintf(psize,24,"%u bytes",atio->GetConfig()->page_size); }
		else
		{  strcpy(psize,"[no pages]");  }
		printf("Detected device: (0x%02x%02x%02x)\n"
		       "  Device name: %-12s    Flash size.: %6u bytes  [0x%x words]\n"
		       "  Vendor.....: %-12s    EEPROM size: %6u bytes\n",
		       dev_sig._vendor,dev_sig._flash_size,dev_sig._device,
		       dev_sig.device,
		       dev_sig.flash_size*2,dev_sig.flash_size,
		       dev_sig.vendor,
		       dev_sig.eeprom_size);
		printf("  Page size: %s;  Fuse+Lock: %u bytes;  ChipErase: %s\n",
		       psize,
		       atio->GetConfig()->fuse_lock_bytes,
		       atio->GetConfig()->has_erase_busy ? "RDY/BSY" : "delay");
	}
	
	current_mode=1;
	return(0);
}

// Always returns 0 and 1 on failure. 
static int EnsureIdleMode()
{
	if(current_mode==0)
	{
		// Seems we're already in idle mode. 
		// Nevertheless, make sure to set proper levels: 
		if(atio)
		{  atio->Disconnect();  }
		return(0);
	}
	
	atio->LeaveProgMode(&caught_signal);
	atio->Disconnect();
	
	current_mode=0;
	return(0);
}


static void _ReadFuseAndLock()
{
	/* NOTE: UNTESTED CODE. Never read sth different than 0xff. */
	
	printf("Reading fuse and lock bits... ");  fflush(stdout);
	uchar buf[8];
	int size=atio->ReadFuseAndLock(buf);
	if(!size) return;
	printf("OK\n");
	
	const char *up="unprog";
	const char *pp="PROG";
	switch(atio->GetConfig()->device)
	{
		case AtmelIO::AD_AT90S2313:
		{
			assert(size==1);
			uchar x=buf[0];
			printf(
				"  Lock bit 1: %-6s      Lock bit 2: %s\n"
				"  SPIEN fuse: %-6s      FSTRT fuse: %s\n",
				(x & (1<<7)) ? up : pp, (x & (1<<6)) ? up : pp,
				(x & (1<<5)) ? up : pp, (x & (1<<0)) ? up : pp);

			if(!(x & (1<<5)))
			{  printf("No Data downloading (SPIEN fuse programmed).\n");  }
		}	break;
		case AtmelIO::AD_AT90S4433:
		{
			assert(size==2);
			uchar fuse=buf[0];
			uchar lock=buf[1];
			printf(
				"  Lock bits: (0x%02x)\n"
				"    Lock bit 1,2: %s %s\n"
				"  Fuse bits: (0x%02x)\n"
				"    CKSEL 0..2: %s %s %s\n"
				"    BODEN: %s    BODLEVEL: %s   SPIEN: %s\n",
				lock,
				 (lock & 0x02) ? up : pp,(lock & 0x04) ? up : pp,
				fuse,
				 (fuse & 0x01) ? up : pp,(fuse & 0x02) ? up : pp,
				 	(fuse & 0x04) ? up : pp,
				 (fuse & 0x08) ? up : pp,(fuse & 0x10) ? up : pp,
				 	(fuse & 0x20) ? up : pp );
		}	break;
		case AtmelIO::AD_ATMega161:
		{
			assert(size==2);
			uchar fuse=buf[0];
			uchar lock=buf[1];
			printf(
				"  Lock bits: (0x%02x)\n"
				"    Lock bit 1,2: %s %s\n"
				"    Boot lock 01,02: %s %s\n"
				"    Boot lock 11,12: %s %s\n"
				"  Fuse bits: (0x%02x)\n"
				"    CKSEL 0..2: %s %s %s\n"
				"    BOOTRST: %s    SPIEN: %s   SUT: %s\n",
				lock,
				 (lock & 0x01) ? up : pp,(lock & 0x02) ? up : pp,
				 (lock & 0x04) ? up : pp,(lock & 0x08) ? up : pp,
				 (lock & 0x10) ? up : pp,(lock & 0x20) ? up : pp,
				fuse,
				 (fuse & 0x01) ? up : pp,(fuse & 0x02) ? up : pp,
				 	(fuse & 0x04) ? up : pp,
				 (fuse & 0x10) ? up : pp,(fuse & 0x20) ? up : pp,
				 	(fuse & 0x40) ? up : pp );
		}	break;
		case AtmelIO::AD_ATMega8515:
		{
			assert(size==3);
			uchar fuse=buf[0];
			uchar lock=buf[1];
			uchar fuseH=buf[2];
			printf(
				"  Lock bits: (0x%02x)\n"
				"    Lock bit 1,2: %s %s\n"
				"    Boot lock 01,02: %s %s\n"
				"    Boot lock 11,12: %s %s\n"
				"  Fuse bits (L:H): (0x%02x:%02x)\n"
				"    CKSEL 0..3: %s %s %s %s\n"
				"    SUT 0,1: %s %s\n"
				"    BOD EN,LEVEL: %s %s\n"
				"    BOOT RST: %s    SZ0,SZ1: %s %s\n"
				"    EESAVE: %s   CKOPT: %s   SPIEN: %s\n"
				"    WDTON: %s   C8515C: %s\n",
				lock,
				 (lock & 0x01) ? up : pp,(lock & 0x02) ? up : pp,
				 (lock & 0x04) ? up : pp,(lock & 0x08) ? up : pp,
				 (lock & 0x10) ? up : pp,(lock & 0x20) ? up : pp,
				fuse,fuseH,
				 (fuse & 0x01) ? up : pp,(fuse & 0x02) ? up : pp,
				 	(fuse & 0x04) ? up : pp,(fuse & 0x08) ? up : pp,
				 (fuse & 0x10) ? up : pp,(fuse & 0x20) ? up : pp,
				 (fuse & 0x40) ? up : pp,(fuse & 0x80) ? up : pp,
				 (fuseH & 0x01) ? up : pp,
					(fuseH & 0x02) ? up : pp,(fuseH & 0x04) ? up : pp,
				 (fuseH & 0x08) ? up : pp,(fuseH & 0x10) ? up : pp,
					(fuseH & 0x20) ? up : pp,
				 (fuseH & 0x40) ? up : pp,(fuseH & 0x80) ? up : pp );
		}	break;
		default:  assert(0);  break;
	}
}


static void _ChipErase()
{
	printf("Performing chip erase... ");  fflush(stdout);
	int rv=atio->ChipErase();
	
	if(rv)
	{
		// Must leave (and enter) programming mode again. 
		EnsureIdleMode();
	}
	else
	{  printf("OK\n");  }
}


static void _DoCompare(DataBuffer *dbA,DataBuffer *dbB)
{
	printf("Comparing buffers \"%s\", \"%s\": ",dbA->name,dbB->name);
	fflush(stdout);
	char *paend=dbA->buf+dbA->size;
	char *pbend=dbB->buf+dbB->size;
	size_t difference=0;
	int dump_diff=0;
	for(char *pa=dbA->buf,*pb=dbB->buf; pa<paend && pb<pbend; pa++,pb++)
	{
		if(*pa!=*pb)
		{  ++difference;  }
	}
	if(!difference && dbA->size==dbB->size)
	{  printf("EQUAL (%u bytes)\n",dbA->size);  }
	else if(difference && dbA->size==dbB->size)
	{  printf("%u/%u bytes differ\n",difference,dbA->size);  dump_diff=1;  }
	else if(!difference && dbA->size!=dbB->size)
	{  printf("euqal until \"%s\" EOB (%u bytes)\n",
		dbA->size<dbB->size ? dbA->name : dbB->name,
		dbA->size<dbB->size ? dbA->size : dbB->size);  }
	else
	{  printf("%u/%u bytes differ until \"%s\" EOB\n",difference,
		dbA->size<dbB->size ? dbA->size : dbB->size,
		dbA->size<dbB->size ? dbA->name : dbB->name);  dump_diff=1;  }
	
	// Write differences: 
	if(dump_diff)
	{
		printf("Differences with byte offset [word offset] page:off delta:\n");
		size_t i=0,last_diff=-1;
		for(char *pa=dbA->buf,*pb=dbB->buf; pa<paend && pb<pbend; pa++,pb++,i++)
		{
			if(*pa==*pb)  continue;
			printf("  0x%04x:  0x%02x <-> 0x%02x  (%c <-> %c)  [0x%04x%c]  "
				"0x%02x:%02x%c %3u\n",
				i,(unsigned int)*(uchar*)pa,(unsigned int)*(uchar*)pb,
				isprint(*pa) ? *pa : '.',isprint(*pb) ? *pb : '.',
				i/2,(i & 1) ? 'H' : 'L',
				(i/2)>>6,(i/2)&0x3fU,(i & 1) ? 'H' : 'L',
				i-last_diff);
			last_diff=i;
		}
	}
}


int main(int argc,char **arg)
{
	// First, scan the command line for --help and --version: 
	for(int i=1; i<argc; i++)
	{
		if(*arg[i]!='-')  continue;
		if(!strcmp(arg[i],"--help") || !strcmp(arg[i],"-help"))
		{  PrintHelp();  return(0);  }
		if(!strcmp(arg[i],"--version"))
		{  printf("%s\n",VERSION);  return(0);  }
	}
	
	// Okay, check download mode: 
	const char *prg_name=strchr(arg[0],'/');
	if(prg_name)  ++prg_name;
	else  prg_name=arg[0];
	
	if(!strcmp(prg_name,"atmelprg"))  downloadmode=DL_Parallel;
	else if(!strcmp(prg_name,"atmelspi"))  downloadmode=DL_Serial;
	else
	{  fprintf(stderr,"You must call this program "
		"\"atmelprg\" or \"atmelspi\".\n");  return(1);  }
	
	dblist=new DataBufferList();
	
	InstallSignalHandler();
	
	// Okay, let's do it: 
	for(int i=1; i<argc; i++)
	{
		if(caught_signal)  break;
		
		const char *cmd=arg[i];
		if(*cmd!='-')
		{
			fprintf(stderr,"Expecting command and not \"%s\" "
				"(skipped).\n",arg[i]);
			continue;
		}
		while(*cmd=='-')  ++cmd;
		char *argA=(i+1<argc) ? arg[i+1] : NULL;
		char *argB=(i+2<argc) ? arg[i+2] : NULL;
		if(argA && *argA=='-')  {  argA=NULL;  argB=NULL;  }
		if(argB && *argB=='-')  {  argB=NULL;  }
		if(!strcmp(cmd,"readsig"))
		{
			if(EnsureProgramMode(1))  break;
			// Nothing more to do. 
		}
		else if(!strcmp(cmd,"readfuse"))
		{
			if(EnsureProgramMode())  break;
			
			_ReadFuseAndLock();
		}
		else if(!strcmp(cmd,"load"))
		{
			if(!argA)
			{  fprintf(stderr,"Command -load needs >=1 argument.\n");
				continue;  }
			char *file=argA;  ++i;
			char *bufname=argB;
			if(!bufname)
			{
				bufname=strrchr(file,'/');
				if(bufname) ++bufname;
				else bufname=file;
			}
			else ++i;
			
			DataBuffer *db=dblist->GetDataBuffer(bufname,1);
			db->ReadBinaryFile(file);
		}
		else if(!strcmp(cmd,"save"))
		{
			if(!argB)
			{  fprintf(stderr,"Command -save needs 2 arguments.\n");
				continue;  }
			char *bufname=argA;  ++i;
			char *file=argB;  ++i;
			DataBuffer *db=dblist->GetDataBuffer(bufname,0);
			if(!db)
			{  fprintf(stderr,"-save: no such buffer -- %s\n",
				bufname);  }
			else
			{  db->WriteBinaryFile(file);  }
		}
		else if(!strcmp(cmd,"dump") || 
		        !strcmp(cmd,"dumpasc"))
		{
			int with_ascii=!strcmp(arg[i],"-dumpasc");
			if(!argA)
			{  fprintf(stderr,"Command -dump needs 1 argument.\n");
				continue;  }
			char *bufname=argA;  ++i;
			DataBuffer *db=dblist->GetDataBuffer(bufname,0);
			if(!db)
			{  fprintf(stderr,"-dump: no such buffer -- %s\n",
				bufname);  }
			else
			{  db->HexDumpBuffer(stdout,with_ascii);  }
		}
		else if(!strcmp(cmd,"erase"))
		{
			if(EnsureProgramMode())  break;
			
			_ChipErase();
		}
		else if(!strcmp(cmd,"readflash"))
		{
			char *bufname=argA;
			if(bufname)  ++i;
			else  bufname="flash";
			
			if(EnsureProgramMode())  break;
			
			DataBuffer *db=dblist->GetDataBuffer(bufname,1);
			db->ReadFlash(atio,0,dev_sig.flash_size*2);
		}
		else if(!strcmp(cmd,"writeflash"))
		{
			char *bufname=argA;
			if(bufname)  ++i;
			else  bufname="flash";
			
			if(EnsureProgramMode())  break;
			
			DataBuffer *db=dblist->GetDataBuffer(bufname,0);
			if(!db)
			{  fprintf(stderr,"-writeflash: no such buffer -- %s\n",
				bufname);  }
			else if(dev_sig.flash_size*2<db->size)
			{  fprintf(stderr,"-writeflash: buffer \"%s\" (%u bytes) "
				"does not fit into %u bytes flash.\n",
				bufname,db->size,dev_sig.flash_size*2);  }
			else if(db->size%2)
			{  fprintf(stderr,"-writeflash: buffer \"%s\" (%u bytes) "
				"has odd size.\n",bufname,db->size);  }
			else
			{  db->WriteFlash(atio,0);  }
		}
		else if(!strcmp(cmd,"readeeprom"))
		{
			char *bufname=argA;
			if(bufname)  ++i;
			else  bufname="eeprom";
			
			if(EnsureProgramMode())  break;
			
			DataBuffer *db=dblist->GetDataBuffer(bufname,1);
			db->ReadEEPROM(atio,0,dev_sig.eeprom_size);
		}
		else if(!strcmp(cmd,"writeeeprom") || 
		        !strcmp(cmd,"writeeprom"))
		{
			char *bufname=argA;
			if(bufname)  ++i;
			else  bufname="eeprom";
			
			if(EnsureProgramMode())  break;
			
			DataBuffer *db=dblist->GetDataBuffer(bufname,0);
			if(!db)
			{  fprintf(stderr,"-writeeeprom: no such buffer -- %s\n",
				bufname);  }
			else if(dev_sig.eeprom_size<db->size)
			{  fprintf(stderr,"-writeeeprom: buffer \"%s\" (%u bytes) "
				"does not fit into %u bytes EEPROM.\n",
				bufname,db->size,dev_sig.eeprom_size);  }
			{  db->WriteEEPROM(atio,0);  }
		}
		else if(!strcmp(cmd,"compare"))
		{
			if(!argB)
			{  fprintf(stderr,"Command -compare needs 2 arguments.\n");
				continue;  }
			char *bufA=argA;  ++i;
			char *bufB=argB;  ++i;
			DataBuffer *dbA=dblist->GetDataBuffer(bufA,0);
			DataBuffer *dbB=dblist->GetDataBuffer(bufB,0);
			if(!dbA || !dbB)
			{  fprintf(stderr,"-compare: no such buffer -- %s\n",
				!dbA ? bufA : bufB);  }
			else
			{  _DoCompare(dbA,dbB);  }
		}
		else if(!strcmp(cmd,"special"))
		{
			// Set the clock source to external for ATMega8515. 
			if(EnsureProgramMode())  break;
			uchar buf[4];
			int size=atio->ReadFuseAndLock(buf);
			assert(size==3);
			fprintf(stderr,"Fuse: L=0x%02x, H=0x%02x\n",buf[0],buf[2]);
			// If any of the bits CKSEL 0..3 is set, clear them. 
			uchar fuseL=buf[0];
			if(!(fuseL & 0x0fU))
			{  fprintf(stderr,"Fuses already set for external clock.\n");  }
			else
			{
				fuseL&=0xf0U;
				fprintf(stderr,"Writing fuse (L) for external clock "
					"(0x%02x).\n",fuseL);
				atio->WriteFuse(fuseL,/*high=*/0);
			}
		}
		else
		{  fprintf(stderr,"Unknown command \"-%s\".\n",cmd);
			continue;  }
	}
	
	EnsureIdleMode();
	
	delete dblist;  dblist=NULL;
	if(atio)
	{  delete atio;  atio=NULL;  }
	if(ppio)
	{  delete ppio;  ppio=NULL;  }
	
	return(0);
}
